@@ -1,22 +1,15 @@
-"""The registration module contains classes for image registration.
-
-Image registration aims to align two images using a particular transformation.
-pymia currently supports multi-modal rigid registration, i.e. align two images of different modalities
-using a rigid transformation (rotation, translation, reflection, or their combination).
+"""The registration module provides classes for image registration.
 
 See Also:
-    `ITK Registration <https://itk.org/Doxygen/html/RegistrationPage.html>`_
-    `ITK Software Guide Registration <https://itk.org/ITKSoftwareGuide/html/Book2/ITKSoftwareGuide-Book2ch3.html>`_
+    - `ITK Registration <https://itk.org/Doxygen/html/RegistrationPage.html>`_
+
+    - `ITK Software Guide Registration <https://itk.org/ITKSoftwareGuide/html/Book2/ITKSoftwareGuide-Book2ch3.html>`_
 """
 import abc
 import enum
 import os
-import typing as t
+import typing
 
-import matplotlib
-matplotlib.use('Agg')  # use matplotlib without having a window appear
-import matplotlib.pyplot as plt
-import numpy as np
 import SimpleITK as sitk
 
 import pymia.filtering.filter as pymia_fltr
@@ -30,11 +23,10 @@ class RegistrationType(enum.Enum):
     BSPLINE = 4
 
 
-class RegistrationCallback(metaclass=abc.ABCMeta):
-    """Represents the abstract handler for the registration callbacks."""
+class RegistrationCallback(abc.ABC):
 
     def __init__(self) -> None:
-        """Initializes a new instance of the abstract RegistrationCallback class."""
+        """Represents the abstract handler for the registration callbacks."""
         self.registration_method = None
         self.fixed_image = None
         self.moving_image = None
@@ -80,12 +72,11 @@ class RegistrationCallback(metaclass=abc.ABCMeta):
         pass
 
 
-class MultiModalRegistrationParams(pymia_fltr.IFilterParams):
-    """Represents parameters for the multi-modal rigid registration."""
+class MultiModalRegistrationParams(pymia_fltr.FilterParams):
 
-    def __init__(self, fixed_image: sitk.Image, fixed_image_mask: sitk.Image=None,
-                 callbacks: t.List[RegistrationCallback]=None):
-        """Initializes a new instance of the MultiModalRegistrationParams class.
+    def __init__(self, fixed_image: sitk.Image, fixed_image_mask: sitk.Image = None,
+                 callbacks: typing.List[RegistrationCallback] = None):
+        """Represents parameters for the multi-modal rigid registration used by the :class:`.MultiModalRegistration` filter.
 
         Args:
             fixed_image (sitk.Image): The fixed image for the registration.
@@ -99,38 +90,26 @@ class MultiModalRegistrationParams(pymia_fltr.IFilterParams):
         self.callbacks = callbacks
 
 
-class MultiModalRegistration(pymia_fltr.IFilter):
-    """Represents a multi-modal image registration filter.
-
-    The filter estimates a 3-dimensional rigid or affine transformation between images of different modalities using
-    - Mutual information similarity metric
-    - Linear interpolation
-    - Gradient descent optimization
-
-    Examples:
-
-    The following example shows the usage of the MultiModalRegistration class.
-
-    >>> fixed_image = sitk.ReadImage('/path/to/image/fixed.mha')
-    >>> moving_image = sitk.ReadImage('/path/to/image/moving.mha')
-    >>> registration = MultiModalRegistration()  # specify parameters to your needs
-    >>> parameters = MultiModalRegistrationParams(fixed_image)
-    >>> registered_image = registration.execute(moving_image, parameters)
-    """
+class MultiModalRegistration(pymia_fltr.Filter):
 
     def __init__(self,
-                 registration_type: RegistrationType=RegistrationType.RIGID,
-                 number_of_histogram_bins: int=200,
-                 learning_rate: float=1.0,
-                 step_size: float=0.001,
-                 number_of_iterations: int=200,
-                 relaxation_factor: float=0.5,
-                 shrink_factors: [int]=(2, 1, 1),
-                 smoothing_sigmas: [float]=(2, 1, 0),
-                 sampling_percentage: float=0.2,
-                 sampling_seed: int=sitk.sitkWallClock,
+                 registration_type: RegistrationType = RegistrationType.RIGID,
+                 number_of_histogram_bins: int = 200,
+                 learning_rate: float = 1.0,
+                 step_size: float = 0.001,
+                 number_of_iterations: int = 200,
+                 relaxation_factor: float = 0.5,
+                 shrink_factors: typing.List[int] = (2, 1, 1),
+                 smoothing_sigmas: typing.List[float] = (2, 1, 0),
+                 sampling_percentage: float = 0.2,
+                 sampling_seed: int = sitk.sitkWallClock,
                  resampling_interpolator=sitk.sitkBSpline):
-        """Initializes a new instance of the MultiModalRegistration class.
+        """Represents a multi-modal image registration filter.
+
+        The filter estimates a 3-dimensional rigid or affine transformation between images of different modalities using
+        - Mutual information similarity metric
+        - Linear interpolation
+        - Gradient descent optimization
 
         Args:
             registration_type (RegistrationType): The type of the registration ('rigid' or 'affine').
@@ -139,14 +118,24 @@ class MultiModalRegistration(pymia_fltr.IFilter):
             step_size (float): The optimizer's step size. Each step in the optimizer is at least this large.
             number_of_iterations (int): The maximum number of optimization iterations.
             relaxation_factor (float): The relaxation factor to penalize abrupt changes during optimization.
-            shrink_factors ([int]): The shrink factors at each shrinking level (from high to low).
-            smoothing_sigmas ([int]):  The Gaussian sigmas for smoothing at each shrinking level (in physical units).
+            shrink_factors (typing.List[int]): The shrink factors at each shrinking level (from high to low).
+            smoothing_sigmas (typing.List[int]):  The Gaussian sigmas for smoothing at each shrinking level (in physical units).
             sampling_percentage (float): Fraction of voxel of the fixed image that will be used for registration (0, 1].
                 Typical values range from 0.01 (1 %) for low detail images to 0.2 (20 %) for high detail images.
                 The higher the fraction, the higher the computational time.
             sampling_seed: The seed for reproducible behavior.
             resampling_interpolator: Interpolation to be applied while resampling the image by the determined
                 transformation.
+
+        Examples:
+
+            The following example shows the usage of the MultiModalRegistration class.
+
+            >>> fixed_image = sitk.ReadImage('/path/to/image/fixed.mha')
+            >>> moving_image = sitk.ReadImage('/path/to/image/moving.mha')
+            >>> registration = MultiModalRegistration()  # specify parameters to your needs
+            >>> parameters = MultiModalRegistrationParams(fixed_image)
+            >>> registered_image = registration.execute(moving_image, parameters)
         """
         super().__init__()
 
@@ -207,11 +196,11 @@ class MultiModalRegistration(pymia_fltr.IFilter):
         self.registration = registration
         self.transform = None
 
-    def execute(self, image: sitk.Image, params: MultiModalRegistrationParams=None) -> sitk.Image:
+    def execute(self, image: sitk.Image, params: MultiModalRegistrationParams = None) -> sitk.Image:
         """Executes a multi-modal rigid registration.
 
         Args:
-            image (sitk.Image): The moving image.
+            image (sitk.Image): The moving image to register.
             params (MultiModalRegistrationParams): The parameters, which contain the fixed image.
 
         Returns:
@@ -287,115 +276,13 @@ class MultiModalRegistration(pymia_fltr.IFilter):
             .format(self=self)
 
 
-class PlotCallback(RegistrationCallback):
-    """Represents a plotter for SimpleITK registrations."""
-
-    def __init__(self, plot_dir: str, file_name_prefix: str='', slice_no: int=-1) -> None:
-        """
-        Args:
-            plot_dir (str): Path to the directory where to save the plots.
-            file_name_prefix (str): The file name prefix for the plots.
-            slice_no (int): The slice number to plot (affects only 3-D images). -1 means to use the middle slice.
-        """
-        super().__init__()
-        self.plot_dir = plot_dir
-        self.file_name_prefix = file_name_prefix
-        self.slice_no = slice_no
-
-        self.metric_values = []
-        self.resolution_iterations = []
-
-    def registration_ended(self):
-        """Callback for the EndEvent."""
-        plt.close()
-
-    def registration_started(self):
-        """Callback for the StartEvent."""
-        self.metric_values = []
-        self.resolution_iterations = []
-
-    def registration_resolution_changed(self):
-        """Callback for the MultiResolutionIterationEvent."""
-        self.resolution_iterations.append(len(self.metric_values))
-
-    def registration_iteration_ended(self):
-        """Callback for the IterationEvent.
-
-        Saves an image including the visualization of the registered images and the metric value plot.
-        """
-
-        self.metric_values.append(self.registration_method.GetMetricValue())
-        # Plot the similarity metric values; resolution changes are marked with a blue star
-        plt.plot(self.metric_values, 'r')
-        plt.plot(self.resolution_iterations, [self.metric_values[index] for index in self.resolution_iterations], 'b*')
-        plt.xlabel('Iteration', fontsize=12)
-        plt.ylabel('Metric Value', fontsize=12)
-
-        # todo(fabianbalsiger): format precision of legends
-        # plt.axes().yaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:2f}'))
-        # plt.axes().xaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:d}'))
-        # _, ax = plt.subplots()
-        # ax.yaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:f2}'))
-
-        # todo(fabianbalsiger): add margin to the left side of the plot
-
-        # Convert the plot to a SimpleITK image (works with the agg matplotlib backend, doesn't work
-        # with the default - the relevant method is canvas_tostring_rgb())
-        plt.gcf().canvas.draw()
-        plot_data = np.fromstring(plt.gcf().canvas.tostring_rgb(), dtype=np.uint8, sep='')
-        plot_data = plot_data.reshape(plt.gcf().canvas.get_width_height()[::-1] + (3,))
-        plot_image = sitk.GetImageFromArray(plot_data, isVector=True)
-
-        # Extract the central axial slice from the two volumes, compose it using the transformation and alpha blend it
-        alpha = 0.7
-
-        moving_transformed = sitk.Resample(self.moving_image, self.fixed_image, self.transform,
-                                           sitk.sitkLinear, 0.0,
-                                           self.moving_image.GetPixelIDValue())
-        # Extract the plotting slice in xy and alpha blend them
-        if self.fixed_image.GetDimension() == 3:
-            slice_index = self.slice_no if self.slice_no != -1 else round((self.fixed_image.GetSize())[2] / 2)
-            image_registration_overlay = (1.0 - alpha) * sitk.Normalize(self.fixed_image[:, :, slice_index]) + \
-                       alpha * sitk.Normalize(moving_transformed[:, :, slice_index])
-        else:
-            image_registration_overlay = (1.0 - alpha) * sitk.Normalize(self.fixed_image) + \
-                       alpha * sitk.Normalize(moving_transformed[:, :])
-
-        combined_slices_image = sitk.ScalarToRGBColormap(image_registration_overlay)
-
-        self._write_combined_image(combined_slices_image, plot_image,
-                                   os.path.join(self.plot_dir,
-                                                self.file_name_prefix + format(len(self.metric_values), '03d') + '.png')
-                                   )
-
-    @staticmethod
-    def _write_combined_image(image1, image2, file_name):
-        """Writes an image including the visualization of the registered images and the metric value plot."""
-        combined_image = sitk.Image(
-            (image1.GetWidth() + image2.GetWidth(), max(image1.GetHeight(), image2.GetHeight())),
-            image1.GetPixelID(), image1.GetNumberOfComponentsPerPixel())
-
-        image1_destination = [0, 0]
-        image2_destination = [image1.GetWidth(), 0]
-
-        if image1.GetHeight() > image2.GetHeight():
-            image2_destination[1] = round((combined_image.GetHeight() - image2.GetHeight()) / 2)
-        else:
-            image1_destination[1] = round((combined_image.GetHeight() - image1.GetHeight()) / 2)
-
-        combined_image = sitk.Paste(combined_image, image1, image1.GetSize(), (0, 0), image1_destination)
-        combined_image = sitk.Paste(combined_image, image2, image2.GetSize(), (0, 0), image2_destination)
-        sitk.WriteImage(combined_image, file_name)
-
-
 class PlotOnResolutionChangeCallback(RegistrationCallback):
-    """Represents a plotter for SimpleITK registrations.
 
-    Saves the moving image on each resolution change and the registration end.
-    """
+    def __init__(self, plot_dir: str, file_name_prefix: str = '') -> None:
+        """Represents a plotter for registrations.
+
+        Saves the moving image on each resolution change and the registration end.
 
-    def __init__(self, plot_dir: str, file_name_prefix: str='') -> None:
-        """
         Args:
             plot_dir (str): Path to the directory where to save the plots.
             file_name_prefix (str): The file name prefix for the plots.
